from os_spage.validator import check_error_reason, simple_check_url


def test_check_error_reason():
    valid_data = ["HTTP 404", "RULE 16", "DNS -2", "SERVER 110", "SSL -2"]
    for data in valid_data:
        assert check_error_reason(data) == True

    invalid_data = ["HTTP404", "RULE  16", "RULE TEST"]
    for data in invalid_data:
        assert check_error_reason(data) == False


def test_simple_check_url():
    valid_data = ["http://www.google.com/", "https://www.google.com/"]
    for data in valid_data:
        assert simple_check_url(data) == True

    invalid_data = ["htp//www.google.com/", "https/www.google.com/"]
    for data in invalid_data:
        assert simple_check_url(data) == False



import abc
import copy
import zlib
from datetime import datetime
from io import BytesIO

from os_rotatefile import open_file

from .common import DEFAULT_ENCODING, TIME_FORMAT
from .compat import StringIO, iteritems
from .default_schema import (
    META_SCHEMA,
    InnerHeaderKeys as I_KEYS,
    RecordTypes as R_TYPES,
    SpageKeys as S_KEYS,
)
from .validator import create_validator


class RecordProcessor(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def process(self, record, **kwargs):
        return record


class SpageRecordProcessor(RecordProcessor):
    def __init__(self, validator, compress=True):
        self._validator = validator
        self._compress = compress

    def _compress_data(self, data):
        return zlib.compress(data)

    def process(self, record, **kwargs):
        if not record[S_KEYS.HTTP_HEADER]:
            record.pop(S_KEYS.HTTP_HEADER)
        inner_header = record[S_KEYS.INNER_HEADER]
        store_size = original_size = (
            len(record[S_KEYS.DATA]) if record[S_KEYS.DATA] is not None else -1
        )

        if original_size >= 0:
            data = record[S_KEYS.DATA]
            r_type = inner_header.get(I_KEYS.TYPE, None)
            if r_type is None:
                if I_KEYS.ORIGINAL_SIZE in inner_header:
                    raise ValueError(
                        "do not specify %s without Type" % I_KEYS.ORIGINAL_SIZE
                    )
                inner_header[I_KEYS.ORIGINAL_SIZE] = original_size
                if self._compress:
                    inner_header[I_KEYS.TYPE] = R_TYPES.COMPRESSED
                    data = self._compress_data(data)
                    store_size = len(data)
                else:
                    inner_header[I_KEYS.TYPE] = R_TYPES.FLAT
            elif r_type == R_TYPES.COMPRESSED or r_type == R_TYPES.DELETED:
                if I_KEYS.ORIGINAL_SIZE not in inner_header:
                    raise ValueError("inner_header require %s" % I_KEYS.ORIGINAL_SIZE)
            elif r_type == R_TYPES.FLAT:
                inner_header[I_KEYS.ORIGINAL_SIZE] = original_size

            record[S_KEYS.DATA] = data
            inner_header[I_KEYS.STORE_SIZE] = store_size
        else:
            if I_KEYS.TYPE not in inner_header:
                inner_header[I_KEYS.TYPE] = R_TYPES.FLAT
            for i in (I_KEYS.ORIGINAL_SIZE, I_KEYS.STORE_SIZE):
                if i in inner_header:
                    inner_header.pop(i)
            record.pop(S_KEYS.DATA)

        self._validator.validate(record)
        return record


class RecordEncoder(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def dumps(self, record, **kwargs):
        raise NotImplementedError


class SpageRecordEncoder(RecordEncoder):
    def __init__(self, allowed_inner_header_keys=None):
        self._inner_header_keys = allowed_inner_header_keys

    def _http_header_str(self, http_header):
        if not http_header:
            return None
        return "\r\n".join(
            [": ".join((k.strip(), v.strip())) for k, v in iteritems(http_header)]
        )

    def _inner_header_str(self, inner_header):
        o = StringIO()
        keys = (
            self._inner_header_keys if self._inner_header_keys else inner_header.keys()
        )
        for k in keys:
            if k in inner_header:
                v = inner_header[k]
                if v is None:
                    continue
                if isinstance(v, datetime):
                    v = v.strftime(TIME_FORMAT)
                o.write(": ".join((str(k).strip(), str(v).strip())))
                o.write("\n")

        data_length = o.tell() - 1
        o.seek(0)
        return o.read(data_length)

    def dumps(self, record, **kwargs):
        o = BytesIO()
        o.write(record[S_KEYS.URL].encode(DEFAULT_ENCODING))
        o.write(b"\n")

        inner_header_str = self._inner_header_str(record[S_KEYS.INNER_HEADER])
        o.write(inner_header_str.encode(DEFAULT_ENCODING))
        o.write(b"\n\n")

        http_header_str = self._http_header_str(record.get(S_KEYS.HTTP_HEADER, None))
        if not http_header_str:
            o.write(b"\r\n")
        else:
            o.write(http_header_str.encode(DEFAULT_ENCODING))
            o.write(b"\r\n\r\n")

        data = record.get(S_KEYS.DATA, None)
        if data is not None:
            o.write(data)
            o.write(b"\r\n")

        o.seek(0)
        return o.read()


class RecordWriter(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def write(self, f, url, inner_header=None, http_header=None, data=None):
        raise NotImplementedError


class SpageRecordWriter(RecordWriter):
    def __init__(self, processor, encoder):
        self._processor = processor
        self._encoder = encoder

    def write(self, f, url, inner_header=None, http_header=None, data=None):
        if not isinstance(data, (bytes, type(None))):
            raise ValueError(
                "bytes-like data is required, not {}".format(type(data).__name__)
            )
        record = {}
        record[S_KEYS.URL] = url
        record[S_KEYS.INNER_HEADER] = (
            {} if inner_header is None else copy.deepcopy(inner_header)
        )
        record[S_KEYS.HTTP_HEADER] = (
            {} if http_header is None else copy.deepcopy(http_header)
        )
        record[S_KEYS.DATA] = data

        record = self._processor.process(record)
        f.write(self._encoder.dumps(record))


def create_writer(**kwargs):
    validator = kwargs.get("validator", None)
    validator = create_validator(META_SCHEMA) if validator is None else validator
    processor = SpageRecordProcessor(validator, kwargs.get("compress", True))
    allowed_keys = validator.schema["properties"][S_KEYS.INNER_HEADER][
        "properties"
    ].keys()
    encoder = SpageRecordEncoder(allowed_keys)
    return SpageRecordWriter(processor, encoder)


class SpageWriter(object):
    def __init__(self, base_filename, roll_size="1G", compress=True, validator=None):
        self._fp = open_file(base_filename, "w", roll_size=roll_size)
        self._record_writer = create_writer(validator=validator, compress=compress)

    def close(self):
        self._fp.close()

    def write(self, url, inner_header=None, http_header=None, data=None, flush=False):
        self._record_writer.write(
            self._fp, url, inner_header=inner_header, http_header=http_header, data=data
        )

        if flush:
            self._fp.flush()


write = create_writer().write



from os_rotatefile import open_file

from .common import COLON, DEFAULT_ENCODING
from .validator import simple_check_url

CONTENT_TYPE = "Content-Type"


def read(fp):
    reader = Reader(fp)
    for record in reader.read():
        yield record


class Reader(object):
    def __init__(self, fp):
        self._fp = fp
        self._url_latest = None
        self._reset()

    def _reset(self):
        self._url = self._url_latest
        self._header = {}
        self._data = {}
        self._read = self._read_header
        self._url_latest = None

    def _generate(self):
        d = {}
        d[u"url"] = self._url
        d[u"header"] = self._header
        d[u"data"] = self._data
        return d

    def _read_header(self):
        line = self._fp.readline()
        if not line:
            raise StopIteration
        try:
            line = line.decode(DEFAULT_ENCODING).strip()
        except Exception:
            return self._read()
        line_length = len(line)
        if line_length <= 0 and self._header:
            self._read = self._read_data
        elif line_length > 1024:
            pass
        elif simple_check_url(line):
            self._reset()
            self._url = line
        else:
            d = line.find(COLON)
            if d > 0:
                key = line[0:d].strip()
                value = line[d + 1 :].strip()
                self._header[key] = value
        return self._read()

    def _split_series(self, series):
        s = [tuple(i.split(",")) for i in series.split(";") if "," in i]
        return s

    def _read_data(self):
        data = {}
        for key, size in self._split_series(self._header[CONTENT_TYPE]):
            size = int(size)
            if size < 0 or self._url_latest is not None:
                return self._generate()
            d = self._fp.read(size)
            if size > 0 and not d:
                raise StopIteration
            data[key] = d

        self._data = data
        return self._generate()

    def read(self):
        while True:
            try:
                yield self._read()
                self._reset()
            except StopIteration:
                return


class OffpageReader(object):
    def __init__(self, base_filename):
        self._fp = open_file(base_filename, "r")

    def close(self):
        self._fp.close()

    def read(self):
        for record in read(self._fp):
            yield record



import zlib

from os_rotatefile import open_file

from .compat import BytesIO
from .validator import simple_check_url


def read(fp):
    reader = Reader(fp)
    for record in reader.read():
        yield record


class Reader(object):
    def __init__(self, fp):
        self._fp = fp
        self._url_latest = None
        self._store_size = 0
        self._reset()

    def _reset(self):
        self._url = self._url_latest
        self._store_size = 0
        self._inner_header = BytesIO()
        self._data = BytesIO()
        self._read = self._read_inner_header
        self._url_latest = None

    def _generate(self):
        out = BytesIO()
        out.write(self._url)
        out.write(b"\n")

        self._inner_header.seek(0)
        out.write(self._inner_header.read())

        self._data.seek(0)
        data = self._data.read()
        original_size = len(data)
        if original_size > 0:
            data = zlib.compress(data)
        store_size = len(data)
        out.write(b"Content-Type: snapshot, %d;\n" % store_size)
        out.write(b"Original-Size: snapshot, %d;\n" % original_size)
        out.write(b"\n")

        out.write(data)
        out.write(b"\n")
        out.seek(0)
        return out.read()

    def _read_inner_header(self):
        line = self._fp.readline()
        if not line:
            raise StopIteration
        line = line.strip()
        line_length = len(line)
        if line_length <= 0 and self._inner_header and self._url:
            self._read = self._read_http_header
        elif line_length > 1024:
            pass
        elif simple_check_url(line):
            self._reset()
            self._url = line
        else:
            if line.startswith(b"Original-Size:"):
                pass
            elif line.startswith(b"Store-Size: "):
                self._store_size = int(line.split(b":")[1].strip())
            else:
                self._inner_header.write(line)
                self._inner_header.write(b"\n")
        return self._read()

    def _read_http_header(self):
        line = self._fp.readline()
        if not line:
            raise StopIteration
        nline = line.strip()
        if not nline:
            self._data.write(line)
            self._read = self._read_data
        elif simple_check_url(nline):
            self._url_latest = nline
            self._read = self._read_data
        else:
            self._data.write(line)
        return self._read()

    def _read_data(self):
        if self._store_size <= 0 or self._url_latest is not None:
            return self._generate()

        data = self._fp.read(self._store_size)
        try:
            data = zlib.decompress(data)
        except:
            pass
        if self._store_size > 0 and not data:
            raise StopIteration

        if (
            self._data.tell() <= 0
        ):  # compat invalid format: no http headers but write two '\r\n'
            if data[0:2] == b"\r\n":
                o = self._fp.read(2)
                data = data[2:] + o
        self._data.write(data)
        return self._generate()

    def read(self):
        while True:
            try:
                yield self._read()
                self._reset()
            except StopIteration:
                return


class SpageToOffpage(object):
    def __init__(self, base_filename):
        self._fp = open_file(base_filename, "r")

    def close(self):
        self._fp.close()

    def read(self):
        for record in read(self._fp):
            yield record



import copy
from datetime import datetime

from jsonschema import Draft4Validator, FormatChecker, validators
from jsonschema.compat import str_types

from .common import TIME_FORMAT
from .compat import iteritems


@FormatChecker.cls_checks("readable_time", raises=ValueError)
def check_datetime(instance):
    if not isinstance(instance, str_types):
        return True
    return datetime.strptime(instance, TIME_FORMAT)


@FormatChecker.cls_checks("url")
def simple_check_url(url):
    b1 = ":"
    b2 = "://"
    if isinstance(url, bytes):
        b1 = b":"
        b2 = b"://"
    if len(url) > 0:
        t = url.find(b1)
        if t > 0 and url[t : t + 3] == b2:
            return True
    return False


ERROR_TYPES = set(["HTTP", "SSL", "RULE", "SERVER", "DNS"])


@FormatChecker.cls_checks("error_reason")
def check_error_reason(err_string):
    c = err_string.split(" ")
    if len(c) != 2:
        return False

    return c[0] in ERROR_TYPES and c[1].lstrip("-").isdigit()


def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in iteritems(properties):
            if "default" in subschema:
                o = subschema["default"]
                if callable(o):
                    o = o()
                instance.setdefault(property, o)

        for error in validate_properties(validator, properties, instance, schema):
            yield error

    return validators.extend(validator_class, {"properties": set_defaults})


DefaultPropertyDraft4Validator = extend_with_default(Draft4Validator)

EXTRA_TYPES = {"datetime": datetime, "bytes": bytes}


def create_validator(schema, extra_types=None, format_checker=None):
    types = copy.deepcopy(EXTRA_TYPES)
    if extra_types:
        types.update(extra_types)
    if format_checker is None:
        format_checker = FormatChecker()

    return DefaultPropertyDraft4Validator(
        schema, types=types, format_checker=format_checker
    )



from os_rotatefile import open_file

from .common import COLON, DEFAULT_ENCODING
from .default_schema import InnerHeaderKeys as I_KEYS
from .validator import simple_check_url


def read(fp):
    reader = Reader(fp)
    for record in reader.read():
        yield record


class Reader(object):
    def __init__(self, fp):
        self._fp = fp
        self._url_latest = None
        self._reset()

    def _reset(self):
        self._url = self._url_latest
        self._inner_header = {}
        self._http_header = {}
        self._data = None
        self._read = self._read_inner_header
        self._url_latest = None

    def _generate(self):
        d = {}
        d[u"url"] = self._url
        d[u"inner_header"] = self._inner_header
        d[u"http_header"] = self._http_header
        d[u"data"] = self._data
        return d

    def _read_inner_header(self):
        line = self._fp.readline()
        if not line:
            raise StopIteration
        try:
            line = line.decode(DEFAULT_ENCODING).strip()
        except Exception:
            return self._read()
        line_length = len(line)
        if line_length <= 0 and self._inner_header and self._url:
            self._read = self._read_http_header
        elif line_length > 1024:
            pass
        elif simple_check_url(line):
            self._reset()
            self._url = line
        else:
            d = line.find(COLON)
            if d > 0:
                key = line[0:d].strip()
                value = line[d + 1 :].strip()
                self._inner_header[key] = value
        return self._read()

    def _read_http_header(self):
        line = self._fp.readline()
        if not line:
            raise StopIteration
        try:
            line = line.decode(DEFAULT_ENCODING).strip()
        except Exception:
            return self._read()
        if not line:
            self._read = self._read_data
        elif simple_check_url(line):
            self._url_latest = line
            self._read = self._read_data
        else:
            d = line.find(COLON)
            if d > 0:
                key = line[0:d].strip()
                value = line[d + 1 :].strip()
                self._http_header[key] = value

        return self._read()

    def _read_data(self):
        size = int(self._inner_header.get(I_KEYS.STORE_SIZE, -1))
        if size < 0 or self._url_latest is not None:
            return self._generate()

        data = self._fp.read(size)
        if size > 0 and not data:
            raise StopIteration

        if (
            not self._http_header
        ):  # compat invalid format: no http headers but write two '\r\n'
            if data[0:2] == b"\r\n":
                o = self._fp.read(2)
                data = data[2:] + o
        self._data = data
        return self._generate()

    def read(self):
        while True:
            try:
                yield self._read()
                self._reset()
            except StopIteration:
                return


class SpageReader(object):
    def __init__(self, base_filename):
        self._fp = open_file(base_filename, "r")

    def close(self):
        self._fp.close()

    def read(self):
        for record in read(self._fp):
            yield record



import pkgutil
import sys

from .offpage_reader import OffpageReader, read as read_offpage
from .spage_reader import SpageReader, read as read_spage
from .spage_to_offpage import SpageToOffpage, read as spage_to_offpage
from .spage_writer import SpageWriter, write as _write

write = _write


def __not_supported_mode(name, **kwargs):
    raise ValueError("mode must be 'r' or 'w'")


def __not_supported_page_type(name, **kwargs):
    raise ValueError("page_type must be 'spage', 'offpage', 's2o'")


def read(s, page_type="spage"):
    r = {"spage": read_spage, "offpage": read_offpage, "s2o": spage_to_offpage}.get(
        page_type, __not_supported_page_type
    )
    return r(s)


def open_file(name, mode, **kwargs):
    r = {
        "w": SpageWriter,
        "r": {"spage": SpageReader, "offpage": OffpageReader, "s2o": SpageToOffpage},
    }.get(mode, __not_supported_mode)
    if mode == "r":
        r = r.get(kwargs.pop("page_type", "spage"), __not_supported_page_type)

    return r(name, **kwargs)


__all__ = ["__version__", "version_info", "open_file"]

__version__ = pkgutil.get_data(__package__, "VERSION").decode("ascii").strip()
version_info = tuple(int(v) if v.isdigit() else v for v in __version__.split("."))

if sys.version_info < (2, 7):
    sys.exit("os-spage %s requires Python 2.7+" % __version__)

del pkgutil



from collections import OrderedDict
from datetime import datetime


class RecordTypes(object):
    FLAT = "flat"
    DELETED = "deleted"
    COMPRESSED = "compressed"


class InnerHeaderKeys(object):
    VERSION = "Version"
    TYPE = "Type"
    FETCH_TIME = "Fetch-Time"
    ORIGINAL_SIZE = "Original-Size"
    STORE_SIZE = "Store-Size"
    BATCH_ID = "batchID"
    ATTACH = "attach"
    IP_ADDRESS = "IP-Address"
    SPIDER_ADDRESS = "Spider-Address"
    DIGEST = "Digest"
    USER_AGENT = "User-Agent"
    FETCH_IP = "Fetch-IP"
    NODE_FETCH_TIME = "Node-Fetch-Time"
    ERROR_REASON = "Error-Reason"


INNER_HEADER_SCHEMA = {
    "type": "object",
    "properties": OrderedDict(
        [
            (InnerHeaderKeys.VERSION, {"type": "string", "default": "1.2"}),
            (
                InnerHeaderKeys.TYPE,
                {  # autofill or specify
                    "type": "string",
                    "enum": set(
                        [
                            getattr(RecordTypes, i)
                            for i in dir(RecordTypes)
                            if not i.startswith("_")
                        ]
                    ),
                },
            ),
            (
                InnerHeaderKeys.FETCH_TIME,
                {  # record store time
                    "anyOf": [
                        {"type": "datetime"},
                        {"type": "string", "format": "readable_time"},
                    ],
                    "default": datetime.now,
                },
            ),
            (
                InnerHeaderKeys.ORIGINAL_SIZE,
                {"type": "number"},  # data(html) size, autofill
            ),
            (InnerHeaderKeys.STORE_SIZE, {"type": "number"}),  # store size, autofill
            (
                InnerHeaderKeys.BATCH_ID,
                {  # batch identity
                    "type": "string",
                    "minLength": 3,
                    "default": "__CHANGE_ME__",
                },
            ),
            (InnerHeaderKeys.ATTACH, {"type": "string"}),
            (
                InnerHeaderKeys.IP_ADDRESS,
                {  # remote host ip
                    "type": "string",
                    "format": "ipv4",
                    "default": "0.0.0.0",
                },
            ),
            (
                InnerHeaderKeys.SPIDER_ADDRESS,
                {"type": "string", "default": "0.0.0.0"},  # spider node identity
            ),
            (
                InnerHeaderKeys.DIGEST,
                {  # can be html md5
                    "type": "string",
                    "default": "0" * 32,
                    "maxLength": 32,
                    "minLength": 32,
                },
            ),
            (InnerHeaderKeys.USER_AGENT, {"type": "string"}),
            (
                InnerHeaderKeys.FETCH_IP,
                {  # generate page machine ip
                    "type": "string",
                    "format": "ipv4",
                    "default": "0.0.0.0",
                },
            ),
            (
                InnerHeaderKeys.NODE_FETCH_TIME,
                {  # real fetch time
                    "anyOf": [
                        {"type": "datetime"},
                        {"type": "string", "format": "readable_time"},
                    ]
                },
            ),
            (
                InnerHeaderKeys.ERROR_REASON,
                {"type": "string", "format": "error_reaseon"},
            ),
        ]
    ),
}


class SpageKeys(object):
    URL = "url"
    INNER_HEADER = "inner_header"
    HTTP_HEADER = "http_header"
    DATA = "data"


META_SCHEMA = {
    "type": "object",
    "properties": {
        SpageKeys.URL: {"type": "string", "format": "url"},
        SpageKeys.INNER_HEADER: INNER_HEADER_SCHEMA,
        SpageKeys.HTTP_HEADER: {"type": "object"},
        SpageKeys.DATA: {"type": "bytes"},
    },
    "required": [SpageKeys.URL, SpageKeys.INNER_HEADER],
}



import zlib

import pytest
from jsonschema import ValidationError

from os_spage import open_file, read, write
from os_spage.compat import iteritems
from os_spage.default_schema import (
    InnerHeaderKeys as I_KEYS,
    RecordTypes as R_TYPES,
    SpageKeys as S_KEYS,
)


def test_write_invalid_data(tmpdir):
    with tmpdir.as_cwd():
        f = open_file("test", "w")
        with pytest.raises(ValidationError):
            f.write(url="abc", inner_header={I_KEYS.BATCH_ID: "test"})


def check_inner_header(w_inner_header, r_inner_header):
    if not w_inner_header:
        assert r_inner_header[I_KEYS.BATCH_ID] == "__CHANGE_ME__"
        return

    for k, v in iteritems(w_inner_header):
        k = str(k)
        if k != I_KEYS.STORE_SIZE:
            assert r_inner_header[k] == str(v)


def check_http_header(w_http_header, r_http_header):
    if w_http_header is None:
        assert r_http_header == {}
        return
    for k, v in iteritems(w_http_header):
        assert r_http_header[str(k)] == str(v)


def check_data(w_data, r_data, r_type):
    if w_data is None:
        assert r_data is None
        return
    if r_type == R_TYPES.COMPRESSED:
        assert zlib.decompress(r_data) == w_data
    else:
        assert r_data == w_data


RECORDS = [
    # inner_header, http_header, data
    (None, None, None),
    ({I_KEYS.BATCH_ID: "test"}, {"k1": "v1"}, b"hello"),
    ({I_KEYS.BATCH_ID: "test"}, {}, b"hello"),
    ({I_KEYS.BATCH_ID: "test"}, {"k1": "v1"}, None),
    ({}, {"k1": "v1"}, b"hello"),
    ({}, {"k1": "v1"}, None),
]


def test_spage_reader_and_writer(tmpdir):
    base_url = "http://www.test.com/"
    filename_prefix = "test_file_"
    with tmpdir.as_cwd():
        f = open_file(filename_prefix, "w", roll_size=100)
        idx = 0
        for inner_header, http_header, data in RECORDS:
            url = base_url + str(idx)
            idx += 1
            f.write(
                url,
                inner_header=inner_header,
                http_header=http_header,
                data=data,
                flush=True,
            )
        f.close()
        f = open_file(filename_prefix, "r")
        idx = 0
        for record in f.read():
            inner_header, http_header, data = RECORDS[idx]
            url = base_url + str(idx)
            idx += 1
            assert url == record[S_KEYS.URL]
            check_inner_header(inner_header, record[S_KEYS.INNER_HEADER])
            check_http_header(http_header, record[S_KEYS.HTTP_HEADER])
            check_data(
                data, record[S_KEYS.DATA], record[S_KEYS.INNER_HEADER][I_KEYS.TYPE]
            )


def test_general_read_and_write(tmpdir):
    base_url = "http://www.test.com/"
    with tmpdir.as_cwd():
        filename = "test_file"
        f = open(filename, "wb")
        idx = 0
        for inner_header, http_header, data in RECORDS:
            url = base_url + str(idx)
            idx += 1
            write(f, url, inner_header=inner_header, http_header=http_header, data=data)
        f.close()
        f = open(filename, "rb")
        idx = 0
        for record in read(f):
            inner_header, http_header, data = RECORDS[idx]
            url = base_url + str(idx)
            idx += 1
            assert url == record[S_KEYS.URL]
            check_inner_header(inner_header, record[S_KEYS.INNER_HEADER])
            check_http_header(http_header, record[S_KEYS.HTTP_HEADER])
            check_data(
                data, record[S_KEYS.DATA], record[S_KEYS.INNER_HEADER][I_KEYS.TYPE]
            )


CUSTOM_RECORDS = [
    # inner_header, http_header, data
    (
        {
            I_KEYS.BATCH_ID: "test",
            I_KEYS.TYPE: R_TYPES.COMPRESSED,
            I_KEYS.ORIGINAL_SIZE: 5,
        },
        {"k1": "v1"},
        zlib.compress(b"hello"),
    ),
    (
        {I_KEYS.BATCH_ID: "test", I_KEYS.TYPE: R_TYPES.FLAT, I_KEYS.ORIGINAL_SIZE: 5},
        {"k1": "v1"},
        b"hello",
    ),
    (
        {
            I_KEYS.BATCH_ID: "test",
            I_KEYS.TYPE: R_TYPES.DELETED,
            I_KEYS.ORIGINAL_SIZE: 5,
        },
        {"k1": "v1"},
        b"hello",
    ),
]


def check_custom_data(w_data, r_data):
    if w_data is None:
        assert r_data is None
        return
    assert r_data == w_data


def test_write_custom_data_and_read(tmpdir):
    base_url = "http://www.test.com/"
    with tmpdir.as_cwd():
        filename_prefix = "test_file_"
        f = open_file(filename_prefix, "w", roll_size=100)
        idx = 0
        for inner_header, http_header, data in CUSTOM_RECORDS:
            url = base_url + str(idx)
            idx += 1
            f.write(url, inner_header=inner_header, http_header=http_header, data=data)
        f.close()
        f = open_file(filename_prefix, "r")
        idx = 0
        for record in f.read():
            inner_header, http_header, data = CUSTOM_RECORDS[idx]
            url = base_url + str(idx)
            idx += 1
            assert url == record[S_KEYS.URL]
            check_inner_header(inner_header, record[S_KEYS.INNER_HEADER])
            check_http_header(http_header, record[S_KEYS.HTTP_HEADER])
            check_custom_data(data, record[S_KEYS.DATA])


INVALID_CUSTOM_RECORDS = [
    # inner_header, http_header, data
    (
        {I_KEYS.BATCH_ID: "test", I_KEYS.TYPE: R_TYPES.COMPRESSED},
        {"k1": "v1"},
        zlib.compress(b"hello"),
    ),
    (
        {I_KEYS.BATCH_ID: "test", I_KEYS.ORIGINAL_SIZE: 5},
        {"k1": "v1"},
        zlib.compress(b"hello"),
    ),
]


def test_write_invalid_custom_data(tmpdir):
    base_url = "http://www.test.com/"
    with tmpdir.as_cwd():
        filename_prefix = "test_file_"
        f = open_file(filename_prefix, "w", roll_size=100)
        idx = 0
        for inner_header, http_header, data in INVALID_CUSTOM_RECORDS:
            url = base_url + str(idx)
            idx += 1
            with pytest.raises(ValueError):
                f.write(
                    url, inner_header=inner_header, http_header=http_header, data=data
                )
        f.close()


def test_read_no_http_headers():
    from io import BytesIO
    from zlib import compress, decompress

    data = b"hello world"
    c_data = compress(data)
    original_size = len(data)
    store_size = len(c_data)
    o = BytesIO()
    o.write(
        b"http://example.com/\nOriginal-Size: %d\nStore-Size: %d\n\n\r\n\r\n"
        % (original_size, store_size)
    )
    o.write(c_data)
    o.seek(0)
    page = next(read(o))
    assert decompress(page["data"]) == data



import operator
import sys

PY3 = sys.version_info[0] >= 3

if PY3:
    from io import StringIO as _StringIO
    from io import BytesIO as _BytesIO

    iteritems = operator.methodcaller("items")


else:
    from StringIO import StringIO as _StringIO
    from StringIO import StringIO as _BytesIO

    iteritems = operator.methodcaller("iteritems")

StringIO = _StringIO
BytesIO = _BytesIO



from io import BytesIO

from os_spage import open_file, read


def test_read_offpage(tmpdir):
    url = "http://www.google.com/"
    data1 = "1" * 10
    data2 = "2" * 11
    data3 = "3" * 12
    raw = """
{url}
Key1: Value1
Content-Type: A, 10;B, 11;C, 12;

{data1}{data2}{data3}
    """.format(
        url=url, data1=data1, data2=data2, data3=data3
    )
    f = tmpdir.join("testfile.dat")
    f.write(raw.encode("utf8"))
    reader = open_file(f.strpath, "r", page_type="offpage")
    for page in reader.read():
        assert page["url"] == "http://www.google.com/"
        assert page["header"]["Key1"] == "Value1"
        assert page["data"]["A"] == data1.encode()
        assert page["data"]["B"] == data2.encode()
        assert page["data"]["C"] == data3.encode()

    s = BytesIO(raw.encode("utf8"))
    for page in read(s, page_type="offpage"):
        assert page["url"] == "http://www.google.com/"
        assert page["header"]["Key1"] == "Value1"
        assert page["data"]["A"] == data1.encode()
        assert page["data"]["B"] == data2.encode()
        assert page["data"]["C"] == data3.encode()
